/*
 	File:		GXtoPostScript.h
 
 	Contains:	QuickDraw GX to PostScript conversion code.
 
 	Version:	Technology:	Quickdraw GX 1.1.x
 	 
 	Copyright:	 1997 by Apple Computer, Inc., all rights reserved.
*/

#ifndef __GXTOPOSTSCRIPT__
#define __GXTOPOSTSCRIPT__


#include "GXToPSBuildConfig.h"
#include "GXPrintingUniverse.h"
#include "PublicPostScriptIE.h"
#include <GXGraphics.h>
#include <GXFonts.h>




/*****************

	This class is used to do translation of
	QuickDraw GX Data to PostScript.
	
	This class defines a THIN layer on top of the 
	good old GX Printing code for the imaging system.
	This class uses Try/Catch for error handling, but all 
	of the underlying code is unchanged in its error handling.
	Errors are propogated up the call chain as the function results.

	The constructor for this class initializes the underlying imaging engine.
	The Destructor shuts it down.
	
	Some notes on font handling:
		If the Font Database is nil,  then font handling will be done on a per DrawShape
		basis.  This is useful for single pass when we didn't have a chance to gather a font database.  If a font database had
		been built, the set it with this.
	
******************/
class CGXtoPostScriptTranslator {

	public:
		CGXtoPostScriptTranslator(CGXtoPostScriptDevice* psDevice, gxPostScriptImageDataRec *imageData, TFontDbase fontDbase, 
															gxMapping *devMapping, gxFormatHalftoneInfo *halftoneRecord);
		~CGXtoPostScriptTranslator();

		void 				DownloadProcSets();
		void 				InitGraphics(gxPostScriptImageDataRec *imageData, gxMapping *devMapping, gxFormatHalftoneInfo *halftoneRecord);
		void				RestoreGraphics();
		void				DrawShape(gxShape theShape, Boolean imageTransferModes = false);
		
	protected:
	
		CGXtoPostScriptDevice			*mDevice;
		TPSIEContext						mPSIEContext;
		TFontDbase							mDbase;
		TFontHandlerContext				mFontHandler;
		scalerStreamTypeFlag			mLegalStreamTypes;
		TFHDownloadFlags					mFHDownloadFlags;	
		long										mVMAvailable;		

		// This function creates the font handler context.				
		void										CreateFontHandler(scalerStreamTypeFlag legalStreamTypes, TFHDownloadFlags flags);

		// Stuff used for walking pictures.   As we traverse down the picture heirarchy, we push and pop transforms.
		//	The stack is then used as the input to PostScriptDrawShape.
		gxShape								InternalDrawShape(gxShape theShape,  Boolean imageTransferModes = false);	// this one does the recursive walk., returns a shape containing shapes with t-modes.
		void										PushTransform(gxTransform tr);
		void										PopTransform();
		gxTransform*						mTransformStack;				// Used for walking picture heirarchies.
		long										mTransformCount;				// How many in stack?
		long										mStackAllocation;					// How many spots allocated?
		
		// For rendering transfer mode shapes.
		gxMapping								mDeviceMapping;					// Grabbed during constructor, and during each InitGraphics call.
		gxColorSpace						mDevColorSpace;					// Color space to use for the transfer mode image.
		gxColorProfile						mDevProfile;						// Color profile to use for the transfer mode image.
		
		void										RenderTransferModes(gxShape theShape, gxShape tModeShape);
		
};//CGXtoPostScriptTranslator

// This macro defines the amount of VM used by the full proc-set.  This is just an approximation.
#define kGXtoPSProcSetVM 			50000

#endif
