/*
 	File:		ColorPrimitives.c
 
 	Contains:	QuickDraw GX to PostScript conversion code.
 						File contains routines for translating colors
						of various types into PostScript.
 
 	Version:	Technology:	Quickdraw GX 1.1.x
 	 
 	Copyright:	 1991-1997 by Apple Computer, Inc., all rights reserved.
*/

#include "GXToPSBuildConfig.h"
#include <GXGraphics.h>
#include "GXGraphicsPriv.h"
#include <GXEnvironment.h>
#include "GXToPostScript.h"
#include "IOUtilities.h"
#include "RDUtil.h"
#include "FontHandler.h"
#include "PublicPostScriptIE.h"
#include "private.h"
#include "PSIEResources.h"
#include "GXErrors.h"
#include "ShapeUtilities.h"

#include <CMICCProfile.h>
#include <CMApplication.h>

#ifdef resumeLabel
	#undef resumeLabel
#endif
#define resumeLabel(exception)

//<FF>

/******************************************
	RGBColorPrimitive:
	
	Routine sets the current color in the graphics
	state to the specified RGB color.
	
	The color had best be in RGB space for this
	to work properly.
	
*******************************************/
OSErr RGBColorPrimitive(TIEGlobalsHdl hIEGlobals, register gxRGBColor* theRGB)
	{
		OSErr								status = noErr;
		TRDParams*					pRDParams;
		register gxColor*		pCurrColor;
		TIEGlobalsPtr				pGlobals;
		
		/** Don't bother sending color if it the same as current color. **/
		
		pGlobals = *hIEGlobals;
		pCurrColor = &( pGlobals->gStateNest[pGlobals->gStateDepth].currColor );

		pGlobals->gStateNest[pGlobals->gStateDepth].level2DevSpace = true; /* Now in Level-2 device space */
		pGlobals->gStateNest[pGlobals->gStateDepth].currSpace = gxRGBSpace;
		
		if ( (pCurrColor->space != gxRGBSpace) ||
				 (pCurrColor->element.rgb.red != theRGB->red) ||
				 (pCurrColor->element.rgb.green != theRGB->green) ||
				 (pCurrColor->element.rgb.blue != theRGB->blue) ||
				 (pGlobals->ieStateFlags & eColorOutOfDate)) {
				 
			pCurrColor->element.rgb = *theRGB;					// Save this color in the graphics state.
			pCurrColor->space = gxRGBSpace;
		
			pRDParams = (*hIEGlobals)->pRDParams;
			
			pRDParams->resIndex = kSetRGBColor;
				
			
			status = RDResPrintf(	pRDParams,
														COMPONENTtoFIXED(theRGB->red),
														COMPONENTtoFIXED(theRGB->green),
														COMPONENTtoFIXED(theRGB->blue)
													);
			nrequire(status, failed);
						
		}//end if
		
failed:
		return(status);	
	
	}//RGBColorPrimitve


//<FF>
/******************************************
	GrayColorPrimitive:
	
	Routine sets the current color in the graphics
	state to the specified gray color.
	
	The color had best be in gray space for this
	to work properly.
	
*******************************************/
OSErr GrayColorPrimitive(TIEGlobalsHdl hIEGlobals, register unsigned short theGray)
	{
		OSErr								status = noErr;
		TRDParams*					pRDParams;
		register gxColor*		pCurrColor;
		TIEGlobalsPtr				pGlobals;
		
		/** Don't bother sending color if it the same as current color. **/

		pGlobals = *hIEGlobals;
		pCurrColor = &( pGlobals->gStateNest[pGlobals->gStateDepth].currColor );
		
		pGlobals->gStateNest[pGlobals->gStateDepth].level2DevSpace = true;	 /* Now in Level-2 device space */
		pGlobals->gStateNest[pGlobals->gStateDepth].currSpace = gxGraySpace;

		if ( (pCurrColor->space != gxGraySpace) || 
				 (pCurrColor->element.gray != theGray) || 
				 (pGlobals->ieStateFlags & eColorOutOfDate)) {

			pCurrColor->element.gray = theGray;					// don't bother copying the whole thing.
			pCurrColor->space = gxGraySpace;
		
			pRDParams = (*hIEGlobals)->pRDParams;
			
			pRDParams->resIndex = kSetGray;
			
			nrequire(status = RDResPrintf(	pRDParams, COMPONENTtoFIXED(theGray) ), failed);

			
		}//end if

failed:		
		return(status);	
	
	}//GrayColorPrimitve


//<FF>
/******************************************
	CMYKColorPrimitive:
	
	Routine sets the current color in the graphics
	state to the specified RGB color.
	
	The color had best be in RGB space for this
	to work properly.
	
*******************************************/
OSErr CMYKColorPrimitive(TIEGlobalsHdl hIEGlobals, register gxCMYKColor* theCMYK)
	{
		OSErr								status = noErr;
		TRDParams*					pRDParams;
		register gxColor*		pCurrColor;
		TIEGlobalsPtr				pGlobals;
		
		/** Don't bother sending color if it the same as current color. **/
		
		pGlobals = *hIEGlobals;
		pCurrColor = &( pGlobals->gStateNest[pGlobals->gStateDepth].currColor );

		pGlobals->gStateNest[pGlobals->gStateDepth].level2DevSpace = true; /* Now in Level-2 device space */
		pGlobals->gStateNest[pGlobals->gStateDepth].currSpace = gxCMYKSpace;
		
		if ( (pCurrColor->space != gxCMYKSpace) ||
				 (pCurrColor->element.cmyk.cyan != theCMYK->cyan) ||
				 (pCurrColor->element.cmyk.magenta != theCMYK->magenta) ||
				 (pCurrColor->element.cmyk.yellow != theCMYK->yellow) ||
				 (pCurrColor->element.cmyk.black != theCMYK->black) ||
				 (pGlobals->ieStateFlags & eColorOutOfDate)) {
				 
			pCurrColor->element.cmyk = *theCMYK;					// Save this color in the graphics state.
			pCurrColor->space = gxCMYKSpace;
		
			pRDParams = (*hIEGlobals)->pRDParams;
			
			pRDParams->resIndex = kSetCMYKColor;
				
			
			status = RDResPrintf(	pRDParams,
														COMPONENTtoFIXED(theCMYK->cyan),
														COMPONENTtoFIXED(theCMYK->magenta),
														COMPONENTtoFIXED(theCMYK->yellow),
														COMPONENTtoFIXED(theCMYK->black)
													);
			nrequire(status, failed);
			
		}//end if
		
failed:
		return(status);	
	
	}//CMYKColorPrimitive

//<FF>
/******************************* Level-2 Color Stuff *******************************/


//<FF>
/*********************************************
	ResponseFunctionPrimitive:
	
	Function makes a function for the profile response
	curve and leaves it on the operand stack.
	
***********************************************/
OSErr ResponseFunctionPrimitive(TRDParams *rdParams, short count, unsigned short response[]);
OSErr ResponseFunctionPrimitive(TRDParams *rdParams, short count, unsigned short response[])
	{
		OSErr										status;
		register short					i;
		register gxColorValue		*pResponseValue;
		
		if (count > 0) {
		
			rdParams->resIndex = kNewArray;						// Create an array, leave it on the stack.
			status = RDResPrintf(rdParams, count);
			nrequire(status, failed_NewArray);

			/* Dump the response data */
			
			rdParams->resIndex = kPutFixed;
			pResponseValue = response;
			for (i = count - 1; i != -1; --i) {
			
				status = RDResPrintf(rdParams, COMPONENTtoFIXED(*pResponseValue++) );
				nrequire(status, failed_Put);
			
			}//end for
	
			/** We left an extra copy of the array index on the stack from PutNext **/
			rdParams->resIndex = kDoPop;
			nrequire(status = RDResPrintf(rdParams), failed_pop);		
			
			rdParams->resIndex = kMakeResponseFunc;
			status = RDResPrintf(rdParams);
			ncheck(status);
			
		} else {
		
			/* Just put an identity procedure on the stack */
			rdParams->resIndex = kIdentityProc;
			status = RDResPrintf(rdParams);
			nrequire(status, failed_Buffer);			
		
		}//end if

failed_Buffer:
failed_pop:		
failed_Put:
failed_NewArray:		
		return(status);
		
	}//ResponseFunctionPrimitive
	
//<FF>
/*********************************************
	CalibratedRGBSpacePrimitive:
	
	Function makes Level-2 color space to
	be for a Calibrated RGB Color Space using the profile
	tristimulus values for red, green, and blue as well
	as the profile ramps to linearize them.
	color space array is left on operand stack.

	theSpace:				Color space to use.
	theProfile:			Color profile to use.
	changed:				(returned)  true if the color space changed from the current state.	

	Will use default profile if nil or zero length is passed in.
	
**********************************************/
OSErr	CalibratedRGBSpacePrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile)
	{
		OSErr											status;
		TRDParams*								pRDParams;
		unsigned short						*responseData;
		CMProfile									*pProfile;
		CMProfileResponse					*responseTables;
		CMHeader									*profileHeader;
		CMProfileChromaticities		*chromaticities;
		
		/* Make sure we have a valid profile */

		if ((theProfile == nil) || (GXGetColorProfile(theProfile, nil) == 0))
			theProfile = GXGetDefaultColorProfile();

		pRDParams = (*hIEGlobals)->pRDParams;
		
		/** Get a pointer to the color profile data **/
		
		GXLockColorProfile(theProfile);
		nrequire(status = GXGetGraphicsError(nil), failed_lockProfile);
		pProfile = (CMProfile*)GXGetColorProfileStructure(theProfile, nil);
		
		/* Get pointers into the profile */
		
		profileHeader = &(pProfile->header);
		responseTables = &(pProfile->response);
		chromaticities = &(pProfile->profile);
				
		/** Output the color profile responses (blue, green, red) **/
		
		/* Point to blue */
		responseData = responseTables->data + responseTables->counts[cmGrayResponse];
																				+ responseTables->counts[cmRedResponse] 
																				+ responseTables->counts[cmGreenResponse];
		status = ResponseFunctionPrimitive(pRDParams, responseTables->counts[cmBlueResponse], responseData);
		nrequire(status, failed_Blue);
		
		/* point to green */
		responseData = responseTables->data + responseTables->counts[cmGrayResponse];
																					responseTables->counts[cmRedResponse];
		status = ResponseFunctionPrimitive(pRDParams, responseTables->counts[cmGreenResponse], responseData);
		nrequire(status, failed_Green);
		
		/* point to red */
		responseData = responseTables->data + responseTables->counts[cmGrayResponse];
		status = ResponseFunctionPrimitive(pRDParams, responseTables->counts[cmRedResponse], responseData);
		nrequire(status, failed_Red);


		/** Output the red/green/blue/black/white tristimulus points and the call to the PS proc. **/
		
		pRDParams->resIndex = kMakeCalRGBSpace;
		status = RDResPrintf(pRDParams,
													XYZCOMPONENTtoFIXED(chromaticities->red.X),		XYZCOMPONENTtoFIXED(chromaticities->red.Y),		XYZCOMPONENTtoFIXED(chromaticities->red.Z),
													XYZCOMPONENTtoFIXED(chromaticities->green.X),	XYZCOMPONENTtoFIXED(chromaticities->green.Y),	XYZCOMPONENTtoFIXED(chromaticities->green.Z),
													XYZCOMPONENTtoFIXED(chromaticities->blue.X), 	XYZCOMPONENTtoFIXED(chromaticities->blue.Y), 	XYZCOMPONENTtoFIXED(chromaticities->blue.Z),
													XYZCOMPONENTtoFIXED(profileHeader->white.X),	XYZCOMPONENTtoFIXED(profileHeader->white.Y),	XYZCOMPONENTtoFIXED(profileHeader->white.Z),
													XYZCOMPONENTtoFIXED(profileHeader->black.X),	XYZCOMPONENTtoFIXED(profileHeader->black.Y),	XYZCOMPONENTtoFIXED(profileHeader->black.Z) );
		nrequire(status, failed_ColorPoints);
		



failed_ColorPoints:
failed_Red:
failed_Green:
failed_Blue:	

		GXUnlockColorProfile(theProfile);

failed_lockProfile:

		return(status);
		
	}//CalibratedRGBSpacePrimitive

/**********************************************
	XYZColorSpacePrimitive:
	
	Function makes a Level-2 color space to be
	for xyzSpace using the specified profile for the
	white point and the black point.  color space array
	is left on operand stack.
	
	If nil or a zero length profile is passed in, the default profile is used for the
	white and black points.
	
************************************************/
OSErr XYZColorSpacePrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile);
OSErr XYZColorSpacePrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile)
	{
		OSErr								status;
		TRDParams*					pRDParams;
		CMProfile						*pProfile;
		CMHeader						*profileHeader;
		
		if ((theProfile == nil) || (GXGetColorProfile(theProfile, nil) == 0))
			theProfile = GXGetDefaultColorProfile();

		/** Get a pointer to the color profile data **/
		
		GXLockColorProfile(theProfile);
		nrequire(status = GXGetGraphicsError(nil), failed_lockProfile);
		pProfile = (CMProfile*)GXGetColorProfileStructure(theProfile, nil);
		
		/* Get pointers into the header which contains the white and black points */
		
		profileHeader = &(pProfile->header);
		
		pRDParams = (*hIEGlobals)->pRDParams;
		
		/** Output the white point and black point to make the level-2 colorspace **/
		
		pRDParams->resIndex = kMakeXYZSpace;
		status = RDResPrintf(pRDParams, XYZCOMPONENTtoFIXED(profileHeader->white.X),
																		XYZCOMPONENTtoFIXED(profileHeader->white.Y),
																		XYZCOMPONENTtoFIXED(profileHeader->white.Z),
																		XYZCOMPONENTtoFIXED(profileHeader->black.X),
																		XYZCOMPONENTtoFIXED(profileHeader->black.Y),
																		XYZCOMPONENTtoFIXED(profileHeader->black.Z) );
		ncheck(status);		

failed_MakeSpace:
failed_lockProfile:

		GXUnlockColorProfile(theProfile);
		
		return(status);
		
	}//XYZColorSpacePrimitive




//<FF>
/**********************************************

	RenderingIntentPrimitive:
	
	This procedure calls ColorSync to leave a 
	rendering intent name parameter on the operand stack.
	If ColorSync couldn't do it, we stream nothing.

	hIEGlobals:			Imaging engine global context.
	theProfile:			profile for which to stream intent.
	didIt:					returned: if we could stream the intent. (pass nil if you don't care)

***********************************************/
OSErr RenderingIntentPrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile, Boolean *didIt);
OSErr RenderingIntentPrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile, Boolean *didIt)
	{
		OSErr									status = noErr;
		CMProfileRef					theICCProf;

		if (didIt)
			*didIt = false;		
		
		theICCProf = GXGetColorProfileReference(theProfile);

		/* Get the profile version */

		if (theICCProf != nil) {

			CMAppleProfileHeader	profHeader;
		
			status = CMGetProfileHeader(theICCProf, &profHeader);
			nrequire(status, failed_getheader);
			if (profHeader.cm2.profileVersion >= cmCS2ProfileVersion  ) {
			
				Boolean				preferredCMMNotFound;	
				CMFlattenUPP	streamProc = NewCMFlattenProc(ICCProfileStreamProc); 				// note assuming this segment stays locked since it is in call-chain.
				
				/* First ensure that the desired CRD is available */
				
				status = MakeCRDAvailable( hIEGlobals, theICCProf);
				nrequire(status, failed_MakeCRDAvaialble);
	
				/* Call colorsync to stream the PostScript data for the color space array object */
				
				#if DEBUGLEVEL > 0
					dprintf(trace, "Calling colorsync to stream rendering intent, reference: %X", theICCProf);
				#endif
	
				status = CMGetPS2ColorRenderingIntent(theICCProf,  ((*hIEGlobals)->params.renderOptions & gxNeedsHexOption) ? cmPS7bit : cmPS8bit,
																		streamProc, hIEGlobals, &preferredCMMNotFound);
				DisposeRoutineDescriptor(streamProc);
				nrequire(status, failed_cs2Stream);
				
				
				if (didIt)
					*didIt = true;
			
			}//end if

		}//end if

failed_cs2Stream:
failed_MakeCRDAvaialble:
failed_getheader:

		return(status);
		
	}//RenderingIntentPrimitive
	

/**********************************************

	ColorSpaceArrayPrimitive:
	
	Function uses the ColorSync API for streaming a color space array
	or imaging engine's built in code, depending on the type of profile.
	
	hIEGlobals:			Imaging engine global context.
	theProfile:			The color profile object from which to make a Color Space Array.
	theSpace:				The color space we are tyring to make.
	
	returns error code.
	
***********************************************/

OSErr ColorSpaceArrayPrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile, gxColorSpace theSpace);
OSErr ColorSpaceArrayPrimitive(TIEGlobalsHdl hIEGlobals, gxColorProfile theProfile, gxColorSpace theSpace)
	{
		OSErr									status = noErr;
		CMProfileRef					theICCProf;
		Boolean								preferredCMMNotFound;	
		CMAppleProfileHeader	profHeader;
		
		theICCProf = GXGetColorProfileReference(theProfile);

		/* Get the profile version */

		if (theICCProf != nil) {
		
			status = CMGetProfileHeader(theICCProf, &profHeader);
			nrequire(status, failed_getheader);
						
			/* Make sure the GX color space matches the data type of the ICC profile
				  and it is one we know that the ColorSync can stream (v2.1 for GX 1.2).
					If this test fails, then we will set the ICC ref to nil, causing us to
					use our built in. code.
			*/
			
			switch (theSpace) {
			
				case gxXYZSpace:
				case gxXYZ32Space:
					theICCProf = nil;					// ColorSync 2.1 won't stream XYZ color space arrays.
					break;

				case gxRGBSpace:
				case gxRGB32Space:
					if (profHeader.cm2.dataColorSpace != cmRGBData)
						theICCProf = nil;
					break;
					
				case gxGraySpace:
					if (profHeader.cm2.dataColorSpace != cmGrayData)
						theICCProf = nil;
					break;
				
				default:
					theICCProf = nil;
					break;
			
			}//end switch
			
		}//end if		

		/* See if we have a version 2 profile to try streaming usning ColorSync's api. */

		if ( (theICCProf != nil) && (profHeader.cm2.profileVersion >= cmCS2ProfileVersion ) ) {

			CMFlattenUPP	streamProc = NewCMFlattenProc(ICCProfileStreamProc); 				// note assuming this segment stays locked since it is in call-chain.

			/* Call colorsync to stream the PostScript data for the color space array object */
			
			#if DEBUGLEVEL > 0
				dprintf(trace, "Calling colorsync to stream profile, reference: %X", theICCProf);
			#endif

			status = CMGetPS2ColorSpace(theICCProf,  ((*hIEGlobals)->params.renderOptions & gxNeedsHexOption) ? cmPS7bit : cmPS8bit,
																	streamProc, hIEGlobals, &preferredCMMNotFound);
			DisposeRoutineDescriptor(streamProc);
			nrequire(status, failed_cs2Stream);
			
		} else {
		
			/* If we didn't have a reference, or the reference was to a 1.0 profile, use our own code for Color Space Arrays */
			
			switch(theSpace) {
				
				case gxXYZSpace:
				case gxXYZ32Space:
					status = XYZColorSpacePrimitive(hIEGlobals, theProfile);
					nrequire(status, failed_Space);
					break;
				
				case gxRGBSpace:
				case gxRGB32Space:
					status = CalibratedRGBSpacePrimitive(hIEGlobals, theProfile);
					nrequire(status, failed_Space);
					break;
					
				case gxGraySpace:
					#if DEBUGLEVEL > 0
						dprintf(trace, "Should add color space array streaming for gray");
					#endif
					break;
					
				default:
					#if DEBUGLEVEL > 0
						dprintf(notrace, "Bug, unsupported color space in color space primitive: %d", theSpace);
					#endif
					status = cmErrIncompatibleProfile;
					break;
			}//end switch
	
		}//end if

failed_Space:
failed_cs2Stream:
failed_rdFlush:
failed_getheader:

		return(status);
		
	}//ColorSpaceArrayPrimitive


//<FF>
/*********************************************
	ColorSpacePrimitive
	
	Function makes PostScript level-2 color
	space to reflect the colorSpace and the colorProfile
	of the shape color.  If desired, the color space is set 
	in the graphics state, otherwise it is just left on the 
	operand stack.

	theSpace:				Color space to use.
	theProfile:			Color profile to use.
	setIt:					pass true to set the space in the graphics state, false to leave it on the stack.
	changed:				(returned)  true if the color space changed from the current state. (pass nil if you don't care)	

***********************************************/
OSErr ColorSpacePrimitive(TIEGlobalsHdl hIEGlobals, gxColorSpace theSpace, gxColorProfile theProfile,
																	Boolean setIt, Boolean *changed)
	{
		OSErr								status = noErr;
		TIEGlobalsPtr				pGlobals;
		TGstate							*pGstate;
		TRDParams*					pRDParams;
		gxColorProfile			oldProfile;
			
		pGlobals = *hIEGlobals;
		pGstate = &(pGlobals->gStateNest[pGlobals->gStateDepth]);
		pRDParams = pGlobals->pRDParams;	
		
		if ((pGstate->currSpace != theSpace) || 											/* if the space changed */
				 pGstate->level2DevSpace ||																/* or Switching from L2 device space to calibrated */
				!GXEqualColorProfile(pGstate->currProfile, theProfile) || /* or The profile changed */
				(pGlobals->ieStateFlags & eColorOutOfDate) ||							/* or the state is out of date */
				!setIt) {																									/* or we are just supposed put the space on the stack */

			/************
				If we are supposed to set it in the graphics state, fix our internal state as well.
			*************/
			
			if (setIt) {
			
				/******
					Warning:  cloning the color profile will not work if the profile is a system one
					such as that returned by GetDefaultColorProfile.  This, however will not be the
					case for data that has been unflattened.  If we choose to make the Imaging Engine
					work as an API call, then this could be a problem if apps hand use colors that
					have default or system profiles on them
				*******/
				pGstate->currSpace = theSpace;
				oldProfile = pGstate->currProfile;
				pGstate->currProfile = GXCloneColorProfile(theProfile);		/* watch out if graphics moves mem, pGstate points into handle */
				GXDisposeColorProfile(oldProfile);
				if (changed != nil)
					*changed = true;
					
			}//end if


			/* Stream a color space array for the currentprofile */
			
			status = ColorSpaceArrayPrimitive( hIEGlobals, theProfile, theSpace);
			nrequire(status, failed_Space);
			
			/* Stream a rendering intent for the current profile */

			status = RenderingIntentPrimitive( hIEGlobals, theProfile, nil);
			nrequire(status, failed_Intent);
			
			/** Set the color space if we are supposed to, otherwise it is just left on the stack **/
			
			
			/** Now make the GX Color Space Dictionary **/
			
			pRDParams->resIndex = MakeCSDict;
			status = RDResPrintf(pRDParams);
			nrequire(status, failed_makeCSDict);
			
			if (setIt) {
			
				pRDParams->resIndex = kSetColorSpace;
				status = RDResPrintf(pRDParams);
				ncheck(status);
				
				pGlobals = *hIEGlobals;
				pGstate = &(pGlobals->gStateNest[pGlobals->gStateDepth]);
				pGstate->level2DevSpace = false;			/* current graphics state is in calibrated mode now */
				
			}//end if
			
		} else {
			
			if (changed != nil)
				*changed = false;
		
		}//end if
		
failed_makeCSDict:
failed_Intent:
failed_Space:	
		return(status);
	
	}//ColorSpacePrimitive



//<FF>
/*********************************************
	Level2XYZRGBColorPrimitive:
	
	Output a level-2 xyz Color.
	This routine works for RGB or XYZ space.
	
**********************************************/
OSErr	Level2XYZRGBColorPrimitive(TIEGlobalsHdl hIEGlobals, gxColor *theColor);
OSErr	Level2XYZRGBColorPrimitive(TIEGlobalsHdl hIEGlobals, gxColor *theColor)
	{
		OSErr								status = noErr;
		TIEGlobalsPtr				pGlobals;
		TRDParams*					pRDParams;
		Boolean							spaceChanged;
		gxColor							*pCurrColor;
		
		status = ColorSpacePrimitive(hIEGlobals, theColor->space, theColor->profile, true, &spaceChanged);
		nrequire(status, failed_Space);

		pGlobals = *hIEGlobals;
		pRDParams = pGlobals->pRDParams;	
		pCurrColor = &( pGlobals->gStateNest[pGlobals->gStateDepth].currColor );
		
		/** If the color is different or we are forcing update **/
		if ( spaceChanged ||
				 (pCurrColor->element.component[0] != theColor->element.component[0]) ||
				 (pCurrColor->element.component[1] != theColor->element.component[1]) ||
				 (pCurrColor->element.component[2] != theColor->element.component[2]) ||
				 (pGlobals->ieStateFlags & eColorOutOfDate)) {

			/** Update the current state **/			
			pCurrColor->element.component[0] = theColor->element.component[0];
			pCurrColor->element.component[1] = theColor->element.component[1];
			pCurrColor->element.component[2] = theColor->element.component[2];

			/** Output the PostScript (either RGB or XYZ) **/
			
			if (theColor->space == gxXYZSpace) {			
				
				pRDParams->resIndex = kSetColorF;
				status = RDResPrintf(pRDParams, XYZCOMPONENTtoFIXED(theColor->element.component[0]), 
																				XYZCOMPONENTtoFIXED(theColor->element.component[1]),
																				XYZCOMPONENTtoFIXED(theColor->element.component[2]) );
			
			} else {
			
				pRDParams->resIndex = kSetColorF;
				status = RDResPrintf(pRDParams, COMPONENTtoFIXED(theColor->element.component[0]), 
																				COMPONENTtoFIXED(theColor->element.component[1]),
																				COMPONENTtoFIXED(theColor->element.component[2]) );
			}//end if
			
			nrequire(status, failed_Color);

			
		}//end if
				
failed_Color:
failed_Space:

		return(status);
	
	}//Level2XYZRGBColorPrimitive
	
//<FF>
/**********************************************

	GetLevel2ColorSpace:
	
	This function returns which color space to use
	in Level-2 for the supplied source color space.
	
	sourceSpace:			the color space of the data.
	sourceProfile:		the color profile of the data.
	level2Space:			(returned)  The color space to use for Level-2 device independent colors.
	nComponents:			(returned)  The number of components in the color space
	useDeviceSpace:		(returned)	Whether or not to use a device space.
	
***********************************************/
OSErr	_GetLevel2ColorSpace(gxColorSpace sourceSpace, gxColorProfile sourceProfile, gxColorSpace *level2Space, long *nComponents, Boolean *useDeviceSpace)
	{
		long			profileVersion;
		long			profileSize;
		OSErr			status = noErr;
		
		if (sourceProfile == nil)
			sourceProfile = GXGetDefaultColorProfile();
		
		profileSize = GetColorProfileSize(sourceProfile, &profileVersion);
				
		
		if ( profileSize == 0) 					// means no matching.
			*useDeviceSpace = true;
		else
			*useDeviceSpace = false;
			
		switch(sourceSpace) {
		
			/** For the rgb family, just make it rgb space **/
			
			case gxRGBSpace:
			case gxRGB16Space:
			case gxRGB32Space:
			case gxRGBASpace:
			case gxARGB32Space:
			case gxHSVSpace:
			case gxHSV32Space:
			case gxHLSSpace:
			case gxHLS32Space:
				*level2Space = gxRGBSpace;
				*nComponents = 3;
				break;
			
			/** For the gray family, go to rgb **/

			case gxGrayASpace:
			case gxGraySpace:
				if (*useDeviceSpace) {
					*level2Space = gxGraySpace;
					*nComponents = 1;
				} else {
					*level2Space = gxRGBSpace;
					*nComponents = 3;
				}//end if
				break;
				
			case gxCMYKSpace:
			case gxCMYK32Space:
				if (*useDeviceSpace) {
					*level2Space = gxCMYKSpace;
					*nComponents = 4;
				} else {
					*level2Space = gxXYZSpace;
					*nComponents = 3;
				}//end if
				break;
			
			/** For cieSpace (xyY), convert to xyz **/
			
			case gxYXYSpace:
			case gxYXY32Space:
			case gxYIQSpace:
			case gxYIQ32Space:
			case gxLABSpace:
			case gxLAB32Space:
			case gxLUVSpace:
			case gxLUV32Space:
			case gxXYZSpace:
			case gxXYZ32Space:
				*level2Space = gxXYZSpace;
				*nComponents = 3;
				*useDeviceSpace = false;				// There is no such thing as device XYZ
				break;

			default:
				#if DEBUGLEVEL > 0
					dprintf(notrace, "Bug, don't know colorspace: %d", sourceSpace);
				#endif
				status = colorSpace_out_of_range;
				break;
			
		}//end switch
	
		return(status);
		
	}//GetLevel2ColorSpace


//<FF>
/**********************************************

	Level2ColorPrimitive:
	
		This function generates Level-2 PostScript Device Independent
		color operators for the color.
		
************************************************/
OSErr Level2ColorPrimitive(TIEGlobalsHdl hIEGlobals, gxColor *theColor)
	{
		OSErr								status = noErr;
		gxColorSpace				level2Space;
		long								nComponents;
		Boolean							useDeviceSpace;

		/******
			If the profile is nil Make it the default profile copy so
			we can pass it through clone and dispose
			without headaches
		*******/
		if (theColor->profile == nil)
			theColor->profile = (*hIEGlobals)->defaultProfileCopy;
		
		/** First pass, if the color is an indexed color make it a regular color **/
		
		if (theColor->space == gxIndexedSpace) {
		
			gxColorSpace	theSpace;
			
			GXGetColorSet(theColor->element.indexed.set, &theSpace, nil);
			GXConvertColor(theColor, theSpace, nil, theColor->profile);
			
		}//end if

		/** Second pass, get the color into a space Level-2 can handle **/
		
		status = GetLevel2ColorSpace(theColor->space, theColor->profile, &level2Space, &nComponents, &useDeviceSpace);
		nrequire(status, failed_GetL2Space);
		GXConvertColor(theColor, level2Space, nil, theColor->profile);
				
		/** Now output a level-2 color space and color **/
		/**
			Note, the RGB, Gray, and CMYK color primitives use setrgb and setcmyk which automatically 
					set the current colorspace to /DeviceRGB or /DeviceCMYK or /DeviceGray
		***/
		switch(theColor->space) {
		
			case gxXYZSpace:
				status = Level2XYZRGBColorPrimitive(hIEGlobals, theColor);
				break;
	
			case gxRGBSpace:
				if (useDeviceSpace)
					status = RGBColorPrimitive(hIEGlobals, &(theColor->element.rgb));
				else
					status = Level2XYZRGBColorPrimitive(hIEGlobals, theColor);
				ncheck(status);
				break;
				
			case gxCMYKSpace:
				if (useDeviceSpace) {
					status = CMYKColorPrimitive(hIEGlobals, &(theColor->element.cmyk));
				} else {
					#if DEBUGLEVEL > 1
					dprintf(notrace, "Curious if we get here - cmyk non device level2 space makes no sense");
					#endif
				}//end if
				ncheck(status);					
				break;
				
			case gxGraySpace:
				if (useDeviceSpace) {
					status = GrayColorPrimitive(hIEGlobals, theColor->element.gray);
				} else {
					#if DEBUGLEVEL > 1
					dprintf(notrace, "Curious if we get here - gray non device level2 space is unsupported");
					#endif
					status = colorSpace_out_of_range;
				}//end if
				ncheck(status);
				break;
								
			default:
				#if DEBUGLEVEL > 0
					dprintf(notrace, "Bug, Don't know colorspace: %d", theColor->space);
				#endif
				status = colorSpace_out_of_range;
				break;
					
		}//end switch
		
		
failed_Color:
failed_Space:
failed_GetL2Space:

		return(status);
		
	}//Level2ColorPrimitive